# -*- test-case-name: vertex.test.test_juice -*-
# Copyright 2005 Divmod, Inc.  See LICENSE file for details

__metaclass__ = type

import warnings

from vertex.liner import LineReceiver

from twisted.internet.main import CONNECTION_LOST
from twisted.internet.defer import Deferred, maybeDeferred, fail
from twisted.internet.protocol import ServerFactory, ClientFactory
from twisted.python.failure import Failure
from twisted.python import log

from epsilon import extime

from vertex.sslverify import Certificate, problemsFromTransport, PeerVerifyError

ASK = '_ask'
ANSWER = '_answer'
COMMAND = '_command'
ERROR = '_error'
ERROR_CODE = '_error_code'
ERROR_DESCRIPTION = '_error_description'
LENGTH = '_length'
BODY = 'body'



class JuiceBox(dict):
    """ I am a packet in the JUICE protocol.  """

    def __init__(self, __body='', **kw):
        self.update(kw)
        if __body:
            assert isinstance(__body, str), "body must be a string: %r" % ( repr(__body),)
            self['body'] = __body

    def body():
        def get(self):
            warnings.warn("body attribute of boxes is now just a regular field",
                          stacklevel=2)
            return self['body']
        def set(self, newbody):
            warnings.warn("body attribute of boxes is now just a regular field",
                          stacklevel=2)
            self['body'] = newbody
        return get,set
    body = property(*body())

    def copy(self):
        newBox = self.__class__()
        newBox.update(self)
        return newBox

    def serialize(self,
                  delimiter='\r\n',
                  escaped='\r\n '):
        assert LENGTH not in self

        L = []
        for (k, v) in self.iteritems():
            if k == BODY:
                k = LENGTH
                v = str(len(self[BODY]))
            L.append(k.replace('_', '-').title())
            L.append(': ')
            L.append(v.replace(delimiter, escaped))
            L.append(delimiter)

        L.append(delimiter)
        if BODY in self:
            L.append(self[BODY])

        bytes = ''.join(L)
        return bytes

    def sendTo(self, proto):
        """
        Serialize and send this box to a Juice instance.  By the time it is
        being sent, several keys are required.  I must have exactly ONE of:

            -ask
            -answer
            -error

        If the '-ask' header is set, then the '-command' header must also be
        set.
        """
        proto.sendPacket(self)

# juice.Box => JuiceBox

Box = JuiceBox

class TLSBox(JuiceBox):
    def __init__(self, __certificate, __verify=None, __sslstarted=None, **kw):
        super(TLSBox, self).__init__(**kw)
        self.certificate = __certificate
        self.verify = __verify
        self.sslstarted = __sslstarted

    def sendTo(self, proto):
        super(TLSBox, self).sendTo(proto)
        if self.verify is None:
            proto.startTLS(self.certificate)
        else:
            proto.startTLS(self.certificate, self.verify)
        if self.sslstarted is not None:
            self.sslstarted()

class QuitBox(JuiceBox):
    def sendTo(self, proto):
        super(QuitBox, self).sendTo(proto)
        proto.transport.loseConnection()

class _SwitchBox(JuiceBox):
    def __init__(self, __proto, **kw):
        super(_SwitchBox, self).__init__(**kw)
        self.innerProto = __proto

    def sendTo(self, proto):
        super(_SwitchBox, self).sendTo(proto)
        proto._switchTo(self.innerProto)

class NegotiateBox(JuiceBox):
    def sendTo(self, proto):
        super(NegotiateBox, self).sendTo(proto)
        proto._setProtocolVersion(int(self['version']))

class JuiceError(Exception):
    pass

class RemoteJuiceError(JuiceError):
    """
    This error indicates that something went wrong on the remote end of the
    connection, and the error was serialized and transmitted to you.
    """
    def __init__(self, errorCode, description, fatal=False):
        """Create a remote error with an error code and description.
        """
        Exception.__init__(self, "Remote[%s]: %s" % (errorCode, description))
        self.errorCode = errorCode
        self.description = description
        self.fatal = fatal

class UnhandledRemoteJuiceError(RemoteJuiceError):
    def __init__(self, description):
        errorCode = "UNHANDLED"
        RemoteJuiceError.__init__(self, errorCode, description)

class JuiceBoxError(JuiceError):
    pass

class MalformedJuiceBox(JuiceBoxError):
    pass

class UnhandledCommand(JuiceError):
    pass


class IncompatibleVersions(JuiceError):
    pass

class _Transactor:
    def __init__(self, store, callable):
        self.store = store
        self.callable = callable

    def __call__(self, box):
        return self.store.transact(self.callable, box)

    def __repr__(self):
        return '<Transaction in: %s of: %s>' % (self.store, self.callable)

class DispatchMixin:
    baseDispatchPrefix = 'juice_'
    autoDispatchPrefix = 'command_'

    wrapper = None

    def _auto(self, aCallable, proto, namespace=None):
        if aCallable is None:
            return None
        command = aCallable.command
        if namespace not in command.namespaces:
            # if you're in the wrong namespace, you are very likely not allowed
            # to invoke the command you are trying to invoke.  some objects
            # have commands exposed in a separate namespace for security
            # reasons, since the security model is a role : namespace mapping.
            print 'WRONG NAMESPACE', repr(namespace), repr(command.namespaces)
            return None
        def doit(box):
            kw = stringsToObjects(box, command.arguments, proto)
            for name, extraArg in command.extra:
                kw[name] = extraArg.fromTransport(proto.transport)
#             def checkIsDict(result):
#                 if not isinstance(result, dict):
#                     raise RuntimeError("%r returned %r, not dictionary" % (
#                             aCallable, result))
#                 return result
            def checkKnownErrors(error):
                key = error.trap(*command.allErrors)
                code = command.allErrors[key]
                desc = str(error.value)
                return Failure(RemoteJuiceError(
                        code, desc, error in command.fatalErrors))
            return maybeDeferred(aCallable, **kw).addCallback(
                command.makeResponse, proto).addErrback(
                checkKnownErrors)
        return doit

    def _wrap(self, aCallable):
        if aCallable is None:
            return None
        wrap = self.wrapper
        if wrap is not None:
            return wrap(aCallable)
        else:
            return aCallable

    def normalizeCommand(self, cmd):
        """Return the canonical form of a command.
        """
        return cmd.upper().strip().replace('-', '_')

    def lookupFunction(self, proto, name, namespace):
        """Return a callable to invoke when executing the named command.
        """
        # Try to find a method to be invoked in a transaction first
        # Otherwise fallback to a "regular" method
        fName = self.autoDispatchPrefix + name
        fObj = getattr(self, fName, None)
        if fObj is not None:
            # pass the namespace along
            return self._auto(fObj, proto, namespace)

        assert namespace is None, 'Old-style parsing'
        # Fall back to simplistic command dispatching - we probably want to get
        # rid of this eventually, there's no reason to do extra work and write
        # fewer docs all the time.
        fName = self.baseDispatchPrefix + name
        return getattr(self, fName, None)

    def dispatchCommand(self, proto, cmd, box, namespace=None):
        fObj = self.lookupFunction(proto, self.normalizeCommand(cmd), namespace)
        if fObj is None:
            return fail(UnhandledCommand(cmd))
        return maybeDeferred(self._wrap(fObj), box)

PYTHON_KEYWORDS = [
    'and', 'del', 'for', 'is', 'raise', 'assert', 'elif', 'from', 'lambda',
    'return', 'break', 'else', 'global', 'not', 'try', 'class', 'except',
    'if', 'or', 'while', 'continue', 'exec', 'import', 'pass', 'yield',
    'def', 'finally', 'in', 'print']

def normalizeKey(key):
    lkey = key.lower().replace('-', '_')
    if lkey in PYTHON_KEYWORDS:
        return lkey.title()
    return lkey


def parseJuiceHeaders(lines):
    """
    Create a JuiceBox from a list of header lines.

    @param lines: a list of lines.
    """
    b = JuiceBox()
    bodylen = 0
    key = None
    for L in lines:
        if L[0] == ' ':
            # continuation
            assert key is not None
            b[key] += '\r\n'+L[1:]
            continue
        parts = L.split(': ', 1)
        if len(parts) != 2:
            raise MalformedJuiceBox(L)
        key, value = parts
        key = normalizeKey(key)
        b[key] = value
    return int(b.pop(LENGTH, 0)), b

class JuiceParserBase(DispatchMixin):

    def __init__(self):
        self._outstandingRequests = {}

    def _puke(self, failure):
        log.msg("Juice server or network failure "
                "unhandled by client application:")
        log.err(failure)
        log.msg(
            "Dropping connection!  "
            "To avoid, add errbacks to ALL remote commands!")
        if self.transport is not None:
            self.transport.loseConnection()

    _counter = 0L

    def _nextTag(self):
        self._counter += 1
        return '%x' % (self._counter,)

    def failAllOutgoing(self, reason):
        OR = self._outstandingRequests.items()
        self._outstandingRequests = None # we can never send another request
        for key, value in OR:
            value.errback(reason)

    def juiceBoxReceived(self, box):
        if ANSWER in box:
            question = self._outstandingRequests.pop(box[ANSWER])
            question.addErrback(self._puke)
            self._wrap(question.callback)(box)
        elif ERROR in box:
            question = self._outstandingRequests.pop(box[ERROR])
            question.addErrback(self._puke)
            self._wrap(question.errback)(
                Failure(RemoteJuiceError(box[ERROR_CODE],
                                         box[ERROR_DESCRIPTION])))
        elif COMMAND in box:
            cmd = box[COMMAND]
            def sendAnswer(answerBox):
                if ASK not in box:
                    return
                if self.transport is None:
                    return
                answerBox[ANSWER] = box[ASK]
                answerBox.sendTo(self)
            def sendError(error):
                if ASK not in box:
                    return error
                if error.check(RemoteJuiceError):
                    code = error.value.errorCode
                    desc = error.value.description
                    if error.value.fatal:
                        errorBox = QuitBox()
                    else:
                        errorBox = JuiceBox()
                else:
                    errorBox = QuitBox()
                    log.err(error) # here is where server-side logging happens
                                   # if the error isn't handled
                    code = 'UNHANDLED'
                    desc = "Unhandled Remote System Exception "
                errorBox[ERROR] = box[ASK]
                errorBox[ERROR_DESCRIPTION] = desc
                errorBox[ERROR_CODE] = code
                if self.transport is not None:
                    errorBox.sendTo(self)
                return None # intentionally stop the error here: don't log the
                            # traceback if it's handled, do log it (earlier) if
                            # it isn't
            self.dispatchCommand(self, cmd, box).addCallbacks(sendAnswer, sendError
                                                              ).addErrback(self._puke)
        else:
            raise RuntimeError(
                "Empty packet received over connection-oriented juice: %r" % (box,))

    def sendBoxCommand(self, command, box, requiresAnswer=True):
        """
        Send a command across the wire with the given C{juice.Box}.

        Returns a Deferred which fires with the response C{juice.Box} when it
        is received, or fails with a C{juice.RemoteJuiceError} if an error is
        received.

        If the Deferred fails and the error is not handled by the caller of
        this method, the failure will be logged and the connection dropped.
        """
        if self._outstandingRequests is None:
            return fail(CONNECTION_LOST)
        box[COMMAND] = command
        tag = self._nextTag()
        if requiresAnswer:
            box[ASK] = tag
            result = self._outstandingRequests[tag] = Deferred()
        else:
            result = None
        box.sendTo(self)
        return result






class Argument:
    optional = False

    def __init__(self, optional=False):
        self.optional = optional

    def retrieve(self, d, name):
        if self.optional:
            value = d.get(name)
            if value is not None:
                del d[name]
        else:
            value = d.pop(name)
        return value

    def fromBox(self, name, strings, objects, proto):
        st = self.retrieve(strings, name)
        if self.optional and st is None:
            objects[name] = None
        else:
            objects[name] = self.fromStringProto(st, proto)

    def toBox(self, name, strings, objects, proto):
        obj = self.retrieve(objects, name)
        if self.optional and obj is None:
            # strings[name] = None
            return
        else:
            strings[name] = self.toStringProto(obj, proto)

    def fromStringProto(self, inString, proto):
        return self.fromString(inString)

    def toStringProto(self, inObject, proto):
        return self.toString(inObject)

    def fromString(self, inString):
        raise NotImplementedError()

    def toString(self, inObject):
        raise NotImplementedError()

class JuiceList(Argument):
    def __init__(self, subargs):
        self.subargs = subargs

    def fromStringProto(self, inString, proto):
        boxes = parseString(inString)
        values = [stringsToObjects(box, self.subargs, proto)
                  for box in boxes]
        return values

    def toStringProto(self, inObject, proto):
        return ''.join([objectsToStrings(
                    objects, self.subargs, Box(), proto
                    ).serialize() for objects in inObject])

class ListOf(Argument):
    def __init__(self, subarg, delimiter=', '):
        self.subarg = subarg
        self.delimiter = delimiter

    def fromStringProto(self, inString, proto):
        strings = inString.split(self.delimiter)
        L = [self.subarg.fromStringProto(string, proto)
             for string in strings]
        return L

    def toStringProto(self, inObject, proto):
        L = []
        for inSingle in inObject:
            outString = self.subarg.toStringProto(inSingle, proto)
            assert self.delimiter not in outString
            L.append(outString)
        return self.delimiter.join(L)

class Integer(Argument):
    fromString = int
    def toString(self, inObject):
        return str(int(inObject))

class String(Argument):
    def toString(self, inObject):
        return inObject
    def fromString(self, inString):
        return inString

class EncodedString(Argument):

    def __init__(self, encoding):
        self.encoding = encoding

    def toString(self, inObject):
        return inObject.encode(self.encoding)

    def fromString(self, inString):
        return inString.decode(self.encoding)

# Temporary backwards compatibility for Exponent

Body = String

class Unicode(String):
    def toString(self, inObject):
        # assert isinstance(inObject, unicode)
        return String.toString(self, inObject.encode('utf-8'))

    def fromString(self, inString):
        # assert isinstance(inString, str)
        return String.fromString(self, inString).decode('utf-8')

class Float(Argument):
    fromString = float
    toString = str

class Base64Binary(Argument):
    def toString(self, inObject):
        return inObject.encode('base64').replace('\n', '')
    def fromString(self, inString):
        return inString.decode('base64')

class Time(Argument):
    def toString(self, inObject):
        return inObject.asISO8601TimeAndDate()
    def fromString(self, inString):
        return extime.Time.fromISO8601TimeAndDate(inString)

class ExtraArg:
    def fromTransport(self, inTransport):
        raise NotImplementedError()

class Peer(ExtraArg):
    def fromTransport(self, inTransport):
        return inTransport.getQ2QPeer()

class PeerDomain(ExtraArg):
    def fromTransport(self, inTransport):
        return inTransport.getQ2QPeer().domain

class PeerUser(ExtraArg):
    def fromTransport(self, inTransport):
        return inTransport.getQ2QPeer().resource

class Host(ExtraArg):
    def fromTransport(self, inTransport):
        return inTransport.getQ2QHost()

class HostDomain(ExtraArg):
    def fromTransport(self, inTransport):
        return inTransport.getQ2QHost().domain

class HostUser(ExtraArg):
    def fromTransport(self, inTransport):
        return inTransport.getQ2QHost().resource


class Boolean(Argument):
    def fromString(self, inString):
        if inString == 'True':
            return True
        elif inString == 'False':
            return False
        else:
            raise RuntimeError("Bad boolean value: %r" % (inString,))

    def toString(self, inObject):
        if inObject:
            return 'True'
        else:
            return 'False'

class Command:
    class __metaclass__(type):
        def __new__(cls, name, bases, attrs):
            re = attrs['reverseErrors'] = {}
            er = attrs['allErrors'] = {}
            for v, k in attrs.get('errors',{}).iteritems():
                re[k] = v
                er[v] = k
            for v, k in attrs.get('fatalErrors',{}).iteritems():
                re[k] = v
                er[v] = k
            return type.__new__(cls, name, bases, attrs)

    arguments = []
    response = []
    extra = []
    namespaces = [None]         # This is set to [None] on purpose: None means
                                # "no namespace", not "empty list".  "empty
                                # list" will make your command invalid in _all_
                                # namespaces, effectively uncallable.
    errors = {}
    fatalErrors = {}

    commandType = Box
    responseType = Box

    def commandName():
        def get(self):
            return self.__class__.__name__
            raise NotImplementedError("Missing command name")
        return get,
    commandName = property(*commandName())

    def __init__(self, **kw):
        self.structured = kw
        givenArgs = [normalizeKey(k) for k in kw.keys()]
        forgotten = []
        for name, arg in self.arguments:
            if normalizeKey(name) not in givenArgs and not arg.optional:
                forgotten.append(normalizeKey(name))
#         for v in kw.itervalues():
#             if v is None:
#                 from pprint import pformat
#                 raise RuntimeError("ARGH: %s" % pformat(kw))
        if forgotten:
            if len(forgotten) == 1:
                plural = 'an argument'
            else:
                plural = 'some arguments'
            raise RuntimeError("You forgot %s to %r: %s" % (
                    plural, self.commandName, ', '.join(forgotten)))
        forgotten = []

    def makeResponse(cls, objects, proto):
        try:
            return objectsToStrings(objects, cls.response, cls.responseType(), proto)
        except:
            log.msg("Exception in %r.makeResponse" % (cls,))
            raise
    makeResponse = classmethod(makeResponse)

    def do(self, proto, namespace=None, requiresAnswer=True):
        if namespace is not None:
            cmd = namespace + ":" + self.commandName
        else:
            cmd = self.commandName
        def _massageError(error):
            error.trap(RemoteJuiceError)
            rje = error.value
            return Failure(self.reverseErrors.get(rje.errorCode, UnhandledRemoteJuiceError)(rje.description))

        d = proto.sendBoxCommand(
            cmd, objectsToStrings(self.structured, self.arguments, self.commandType(),
                                  proto),
            requiresAnswer)

        if requiresAnswer:
            d.addCallback(stringsToObjects, self.response, proto)
            d.addCallback(self.addExtra, proto.transport)
            d.addErrback(_massageError)

        return d

    def addExtra(self, d, transport):
        for name, extraArg in self.extra:
            d[name] = extraArg.fromTransport(transport)
        return d


class ProtocolSwitchCommand(Command):
    """Use this command to switch from something Juice-derived to a different
    protocol mid-connection.  This can be useful to use juice as the
    connection-startup negotiation phase.  Since TLS is a different layer
    entirely, you can use Juice to negotiate the security parameters of your
    connection, then switch to a different protocol, and the connection will
    remain secured.
    """

    def __init__(self, __protoToSwitchToFactory, **kw):
        self.protoToSwitchToFactory = __protoToSwitchToFactory
        super(ProtocolSwitchCommand, self).__init__(**kw)

    def makeResponse(cls, innerProto, proto):
        return _SwitchBox(innerProto)

    makeResponse = classmethod(makeResponse)

    def do(self, proto, namespace=None):
        d = super(ProtocolSwitchCommand, self).do(proto)
        proto._lock()
        def switchNow(ign):
            innerProto = self.protoToSwitchToFactory.buildProtocol(proto.transport.getPeer())
            proto._switchTo(innerProto, self.protoToSwitchToFactory)
            return ign
        def die(ign):
            proto.transport.loseConnection()
            return ign
        def handle(ign):
            self.protoToSwitchToFactory.clientConnectionFailed(None, Failure(CONNECTION_LOST))
            return ign
        return d.addCallbacks(switchNow, handle).addErrback(die)

class Negotiate(Command):
    commandName = 'Negotiate'

    arguments = [('versions', ListOf(Integer()))]
    response = [('version', Integer())]

    responseType = NegotiateBox


class Juice(LineReceiver, JuiceParserBase):
    """
    JUICE (JUice Is Concurrent Events) is a simple connection-oriented
    request/response protocol.  Packets, or "boxes", are collections of
    RFC2822-inspired headers, plus a body.  Note that this is NOT a literal
    interpretation of any existing RFC, 822, 2822 or otherwise, but a simpler
    version that does not do line continuations, does not specify any
    particular format for header values, dispatches semeantic meanings of most
    headers on the -Command header rather than giving them global meaning, and
    allows multiple sets of headers (messages, or JuiceBoxes) on a connection.

    All headers whose names begin with a dash ('-') are reserved for use by the
    protocol.  All others are for application use - their meaning depends on
    the value of the "-Command" header.
    """

    protocolName = 'juice-base'

    hostCertificate = None

    MAX_LENGTH = 1024 * 1024

    isServer = property(lambda self: self._issueGreeting,
                        doc="""
                        True if this is a juice server, e.g. it is going to
                        issue or has issued a server greeting upon
                        connection.
                        """)

    isClient = property(lambda self: not self._issueGreeting,
                        doc="""
                        True if this is a juice server, e.g. it is not going to
                        issue or did not issue a server greeting upon
                        connection.
                        """)

    def __init__(self, issueGreeting):
        """
        @param issueGreeting: whether to issue a greeting when connected.  This
        should be set on server-side Juice protocols.
        """
        JuiceParserBase.__init__(self)
        self._issueGreeting = issueGreeting

    def __repr__(self):
        return '<%s %s/%s at 0x%x>' % (self.__class__.__name__, self.isClient and 'client' or 'server', self.innerProtocol, id(self))

    __locked = False

    def _lock(self):
        """ Lock this Juice instance so that no further Juice traffic may be sent.
        This is used when sending a request to switch underlying protocols.
        You probably want to subclass ProtocolSwitchCommand rather than calling
        this directly.
        """
        self.__locked = True

    innerProtocol = None

    def _switchTo(self, newProto, clientFactory=None):
        """ Switch this Juice instance to a new protocol.  You need to do this
        'simultaneously' on both ends of a connection; the easiest way to do
        this is to use a subclass of ProtocolSwitchCommand.
        """

        assert self.innerProtocol is None, "Protocol can only be safely switched once."
        self.setRawMode()
        self.innerProtocol = newProto
        self.innerProtocolClientFactory = clientFactory
        newProto.makeConnection(self.transport)

    innerProtocolClientFactory = None

    def juiceBoxReceived(self, box):
        if self.__locked and COMMAND in box and ASK in box:
            # This is a command which will trigger an answer, and we can no
            # longer answer anything, so don't bother delivering it.
            return
        return super(Juice, self).juiceBoxReceived(box)

    def sendPacket(self, completeBox):
        """
        Send a juice.Box to my peer.

        Note: transport.write is never called outside of this method.
        """
        assert not self.__locked, "You cannot send juice packets when a connection is locked"
        if self._startingTLSBuffer is not None:
            self._startingTLSBuffer.append(completeBox)
        else:
            self.transport.write(completeBox.serialize())

    def sendCommand(self, command, __content='', __answer=True, **kw):
        box = JuiceBox(__content, **kw)
        return self.sendBoxCommand(command, box, requiresAnswer=__answer)

    _outstandingRequests = None
    _justStartedTLS = False

    def makeConnection(self, transport):
        self._transportPeer = transport.getPeer()
        self._transportHost = transport.getHost()
        log.msg("%s %s connection established (HOST:%s PEER:%s)" % (self.isClient and "client" or "server",
                                                                    self.__class__.__name__,
                                                                    self._transportHost,
                                                                    self._transportPeer))
        self._outstandingRequests = {}
        self._requestBuffer = []
        self._sslVerifyProblems = ()
        # ^ Later this will become a mutable list - we can't get the handle
        # during connection shutdown thanks to the fact that Twisted destroys
        # the socket on our transport before notifying us of a lost connection
        # (which I guess is reasonable - the socket is dead by then) See a few
        # lines below in startTLS for details.  --glyph
        LineReceiver.makeConnection(self, transport)

    _startingTLSBuffer = None

    def prepareTLS(self):
        self._startingTLSBuffer = []

    def startTLS(self, certificate, *verifyAuthorities):
        if self.hostCertificate is None:
            self.hostCertificate = certificate
            self._justStartedTLS = True
            self.transport.startTLS(certificate.options(*verifyAuthorities))
            # Remember that mutable list that we were just talking about?  Here
            # it is.  sslverify.py takes care of populating this list as
            # necessary. --glyph
            self._sslVerifyProblems = problemsFromTransport(self.transport)
            stlsb = self._startingTLSBuffer
            if stlsb is not None:
                self._startingTLSBuffer = None
                for box in stlsb:
                    self.sendPacket(box)
        else:
            raise RuntimeError(
                "Previously authenticated connection between %s and %s "
                "is trying to re-establish as %s" % (
                    self.hostCertificate,
                    Certificate.peerFromTransport(self.transport),
                    (certificate, verifyAuthorities)))

    def dataReceived(self, data):
        # If we successfully receive any data after TLS has been started, that
        # means the connection was secured properly.  Make a note of that fact.
        if self._justStartedTLS:
            self._justStartedTLS = False
        return LineReceiver.dataReceived(self, data)

    def connectionLost(self, reason):
        log.msg("%s %s connection lost (HOST:%s PEER:%s)" % (self.isClient and 'client' or 'server',
                                                             self.__class__.__name__,
                                                             self._transportHost,
                                                             self._transportPeer))
        # XXX this may be a slight oversimplification, but I believe that if
        # there are pending SSL errors, they _are_ the reason that the
        # connection was lost.  a totally correct implementation of this would
        # set up a simple state machine to track whether any bytes were
        # received after startTLS was called.  --glyph
        problems = self._sslVerifyProblems
        if problems:
            failReason = Failure(problems[0])
        elif self._justStartedTLS:
            # We just started TLS and haven't received any data.  This means
            # the other connection didn't like our cert (although they may not
            # have told us why - later Twisted should make 'reason' into a TLS
            # error.)
            failReason = PeerVerifyError(
                "Peer rejected our certificate for an unknown reason.")
        else:
            failReason = reason
        self.failAllOutgoing(failReason)
        if self.innerProtocol is not None:
            self.innerProtocol.connectionLost(reason)
            if self.innerProtocolClientFactory is not None:
                self.innerProtocolClientFactory.clientConnectionLost(None, reason)

    def lineReceived(self, line):
        if line:
            self._requestBuffer.append(line)
        else:
            buf = self._requestBuffer
            self._requestBuffer = []
            bodylen, b = parseJuiceHeaders(buf)
            if bodylen:
                self._bodyRemaining = bodylen
                self._bodyBuffer = []
                self._pendingBox = b
                self.setRawMode()
            else:
                self.juiceBoxReceived(b)

    def rawDataReceived(self, data):
        if self.innerProtocol is not None:
            self.innerProtocol.dataReceived(data)
            return
        self._bodyRemaining -= len(data)
        if self._bodyRemaining <= 0:
            if self._bodyRemaining < 0:
                self._bodyBuffer.append(data[:self._bodyRemaining])
                extraData = data[self._bodyRemaining:]
            else:
                self._bodyBuffer.append(data)
                extraData = ''
            self._pendingBox['body'] = ''.join(self._bodyBuffer)
            self._bodyBuffer = None
            b, self._pendingBox = self._pendingBox, None
            self.juiceBoxReceived(b)
            if self.innerProtocol is not None:
                self.innerProtocol.makeConnection(self.transport)
                if extraData:
                    self.innerProtocol.dataReceived(extraData)
            else:
                self.setLineMode(extraData)
        else:
            self._bodyBuffer.append(data)

    protocolVersion = 0

    def _setProtocolVersion(self, version):
        # if we ever want to actually mangle encodings, this is the place to do
        # it!
        self.protocolVersion = version
        return version

    def renegotiateVersion(self, newVersion):
        assert newVersion in VERSIONS, (
            "This side of the connection doesn't support version %r"
            % (newVersion,))
        v = VERSIONS[:]
        v.remove(newVersion)
        return Negotiate(versions=[newVersion]).do(self).addCallback(
            lambda ver: self._setProtocolVersion(ver['version']))

    def command_NEGOTIATE(self, versions):
        for version in versions:
            if version in VERSIONS:
                return dict(version=version)
        raise IncompatibleVersions()
    command_NEGOTIATE.command = Negotiate


VERSIONS = [1]

from cStringIO import StringIO
class _ParserHelper(Juice):
    def __init__(self):
        Juice.__init__(self, False)
        self.boxes = []
        self.results = Deferred()

    def getPeer(self):
        return 'string'

    def getHost(self):
        return 'string'

    disconnecting = False

    def juiceBoxReceived(self, box):
        self.boxes.append(box)

    # Synchronous helpers
    def parse(cls, fileObj):
        p = cls()
        p.makeConnection(p)
        p.dataReceived(fileObj.read())
        return p.boxes
    parse = classmethod(parse)

    def parseString(cls, data):
        return cls.parse(StringIO(data))
    parseString = classmethod(parseString)

parse = _ParserHelper.parse
parseString = _ParserHelper.parseString

def stringsToObjects(strings, arglist, proto):
    objects = {}
    myStrings = strings.copy()
    for argname, argparser in arglist:
        argparser.fromBox(argname, myStrings, objects, proto)
    return objects

def objectsToStrings(objects, arglist, strings, proto):
    myObjects = {}
    for (k, v) in objects.items():
        myObjects[normalizeKey(k)] = v

    for argname, argparser in arglist:
        argparser.toBox(argname, strings, myObjects, proto)
    return strings

class JuiceServerFactory(ServerFactory):
    protocol = Juice
    def buildProtocol(self, addr):
        prot = self.protocol(True)
        prot.factory = self
        return prot

class JuiceClientFactory(ClientFactory):
    protocol = Juice
    def buildProtocol(self, addr):
        prot = self.protocol(False)
        prot.factory = self
        return prot





import os

import pygtk
pygtk.require("2.0")
from twisted.internet import gtk2reactor
gtk2reactor.install()

from twisted.internet import reactor
from twisted.internet.task import LoopingCall

import gtk
import gtk.glade
import egg.trayicon

from vertex.q2qclient import ClientQ2QService
from vertex.q2q import Q2QAddress

class IconAnimator:
    lc = None
    def __init__(self, box, icons):
        self.box = box
        self.icons = icons
        for icon in icons:
            icon.show()
        self.stop()

    def tick(self):
        self.position += 1
        self.position %= len(self.icons)
        c = self.box.get_children()
        if c:
            self.box.remove(c[0])
        self.box.add(self.icons[self.position])

    def start(self):
        self.animating = True
        self.tick()
        self.lc = LoopingCall(self.tick)
        self.lc.start(1.0)

    def stop(self, index=0):
        self.animating = False
        self.position = index - 1
        self.tick()
        if self.lc is not None:
            self.lc.stop()
            self.lc = None


class _NullCb:
    def __init__(self, name):
        self.name = name

    def __call__(self, *a, **kw):
        print 'No callback provided for', self.name, a, kw

class _SignalAttacher:
    def __init__(self, original):
        self.original = original

    def __getitem__(self, callbackName):
        return getattr(self.original, callbackName, None) or _NullCb(callbackName)

GLADE_FILE = os.path.splitext(__file__)[0] + '.glade'


class IdentificationDialog:
    def __init__(self, clientService, notification):
        self.xml = gtk.glade.XML(GLADE_FILE, "ident_dialog")
        self.clientService = clientService
        self.xml.signal_autoconnect(_SignalAttacher(self))
        self.addressEntry = self.xml.get_widget('addressEntry')
        self.passwordEntry = self.xml.get_widget('passwordEntry')
        self.progressBar = self.xml.get_widget('identifyProgressBar')
        self.progressLabel = self.xml.get_widget('identifyProgressLabel')
        self.identifyWindow = self.xml.get_widget("ident_dialog")
        self.cancelButton = self.xml.get_widget('cancelbutton1')
        self.okButton = self.xml.get_widget('okbutton1')
        self.notification = notification

    def identifyCancel(self, event):
        self.identifyWindow.destroy()

    def identifyOK(self, event):
        idstr = self.addressEntry.get_text()
        D = self.clientService.authorize(
            Q2QAddress.fromString(idstr),
            self.passwordEntry.get_text())

        sensitiveWidgets = [self.addressEntry,
                            self.passwordEntry,
                            self.okButton,
                            self.cancelButton]
        for widget in sensitiveWidgets:
            widget.set_sensitive(False)
        self.progressLabel.set_text("Authenticating...")
        def itWorked(workedNone):
            self.identifyWindow.destroy()
            self.notification.setCurrentID(idstr)
        def itDidntWork(error):
            self.progressLabel.set_text(error.getErrorMessage())
            for widget in sensitiveWidgets:
                widget.set_sensitive(True)
        D.addCallbacks(itWorked, itDidntWork)

from twisted.internet.protocol import ServerFactory

class VertexFactory(ServerFactory):

    def __init__(self, notification):
        self.notification = notification

    def startFactory(self):
        self.notification.animator.stop(1)

    def stopFactory(self):
        self.notification.animator.stop(0)



class NotificationEntry:
    def __init__(self):
        self.xml = gtk.glade.XML(GLADE_FILE, "notification_popup")
        icon = egg.trayicon.TrayIcon("Vertex")
        eventbox = gtk.EventBox()
        icon.add(eventbox)
        icon.show_all()

        self.animateItem = self.xml.get_widget("animate")

#         [gtk.image_new_from_icon_name(name, gtk.ICON_SIZE_SMALL_TOOLBAR)
#          for name in ['gnome-mime-text-x-python', 'xchat']]

        icons = [
            gtk.image_new_from_stock(name, gtk.ICON_SIZE_SMALL_TOOLBAR)
            for name in gtk.STOCK_NO, gtk.STOCK_YES
            ]

        self.animator = IconAnimator(eventbox, icons)
        self.xml.signal_autoconnect(_SignalAttacher(self))
        self.menu = self.xml.get_widget("notification_popup")

        self.contactsStart = self.menu.get_children().index(
            self.xml.get_widget("contacts_begin"))

        eventbox.connect('button_press_event', self.popupMenu)
        self.clientService = ClientQ2QService(os.path.expanduser(
                "~/.vertex/q2q-certificates"))

        self.setCurrentID(self.clientService.getDefaultFrom())

    def setCurrentID(self, idName):

        if idName is not None:
            currentID = Q2QAddress.fromString(idName)
            # log in?
            self.animator.start()
            SL = self.xml.get_widget("identifymenuitem").get_children()[0].set_label
            def loggedIn(result):
                SL(str(currentID))
                self.currentID = currentID
            def notLoggedIn(error):
                SL("Identify")
                self.animator.stop(0)
            self.clientService.listenQ2Q(currentID,
                                         {'vertex': VertexFactory(self)},
                                         "desktop vertex UI").addCallbacks(
                loggedIn, notLoggedIn)

    # XXX event handlers

    def toggleAnimate(self, event):
        SL = self.animateItem.get_children()[0].set_label
        if self.animator.animating:
            SL("Animate")
            self.animator.stop()
        else:
            SL("Stop Animating")
            self.animator.start()

    def identifyDialog(self, event):
        IdentificationDialog(self.clientService, self)

    def addContact(self, event):
        mi = gtk.MenuItem("Dummy Contact")
        mi.show_all()
        self.menu.insert(mi, self.contactsStart + 1)

    def popupMenu(self, box, event):
        self.menu.popup(None, None, None, event.button, event.get_time())

    def quit(self, event):
        reactor.stop()


def main():
    import gnome
    gnome.program_init("Vertex", "0.1")
    global ne
    ne = NotificationEntry()
    reactor.run()




"""Utilities and helpers for simulating a network
"""

from cStringIO import StringIO

from vertex.test import utils


def readAndDestroy(iodata):
    try:
        iodata.seek(0)
        result = iodata.read()
        iodata.seek(0)
        iodata.truncate()
    except ValueError:
        print '<bug in FileTransport, early close>'
        result = ''
    return result


class IOPump:
    """Utility to pump data between clients and servers for protocol testing.

    Perhaps this is a utility worthy of being in protocol.py?
    """
    def __init__(self, client, server, clientIO, serverIO, debug):
        self.client = client
        self.server = server
        self.clientIO = clientIO
        self.serverIO = serverIO
        self.debug = debug

    def flush(self, debug=False):
        """Pump until there is no more input or output.

        Returns whether any data was moved.
        """
        result = False
        for x in range(1000):
            if self.pump(debug):
                result = True
            else:
                break
        else:
            assert 0, "Too long"
        return result


    def pump(self, debug=False):
        """Move data back and forth.

        Returns whether any data was moved.
        """
        sData = readAndDestroy(self.serverIO)
        cData = readAndDestroy(self.clientIO)
        self.client.transport._checkProducer()
        self.server.transport._checkProducer()
        if self.debug or debug:
            print '.'
            # XXX slightly buggy in the face of incremental output
            if cData:
                for line in cData.split('\r\n'):
                    print 'C: '+line
            if sData:
                for line in sData.split('\r\n'):
                    print 'S: '+line
        if cData:
            self.server.dataReceived(cData)
        if sData:
            self.client.dataReceived(sData)
        if self.server.transport.disconnecting and not self.server.transport.disconnected:
            if self.debug or debug:
                print '* C'
            self.server.transport.disconnected = True
            self.client.connectionLost(None)
        if self.client.transport.disconnecting and not self.client.transport.disconnected:
            if self.debug or debug:
                print '* S'
            self.client.transport.disconnected = True
            self.server.connectionLost(None)
        return bool(cData or sData)


def connectedServerAndClient(ServerClass, ClientClass,
                             clientTransportWrapper=utils.FileWrapper,
                             serverTransportWrapper=utils.FileWrapper,
                             debug=False):
    """Returns a 3-tuple: (client, server, pump)
    """
    c = ClientClass()
    s = ServerClass()
    cio = StringIO()
    sio = StringIO()
    c.makeConnection(clientTransportWrapper(cio))
    s.makeConnection(serverTransportWrapper(sio))
    pump = IOPump(c, s, cio, sio, debug)
    # kick off server greeting, etc
    pump.flush()
    return c, s, pump



# Copyright 2005 Divmod, Inc.  See LICENSE file for details

import os
from cStringIO import StringIO

from twisted.internet.protocol import FileWrapper
from twisted.internet import defer
from twisted.python.failure import Failure
from twisted.python.filepath import FilePath

from twisted.trial import unittest

from vertex.q2q import Q2QAddress
from vertex import sigma

from vertex.test.iosim import IOPump
from vertex.test.mock_data import data as TEST_DATA

class FakeQ2QTransport(FileWrapper):

    def __init__(self, stringio, q2qhost, q2qpeer):
        FileWrapper.__init__(self, stringio)
        self.q2qhost = q2qhost
        self.q2qpeer = q2qpeer

    def getQ2QPeer(self):
        return self.q2qpeer

    def getQ2QHost(self):
        return self.q2qhost

class FakeDelayedCall:
    def __init__(self, fqs, tup):
        self.fqs = fqs
        self.tup = tup

    def cancel(self):
        self.fqs.calls.remove(self.tup)

class FakeQ2QService:
    # XXX TODO: move this into test_q2q and make sure that all the q2q tests
    # run with it in order to verify that the test harness is not broken.

    def __init__(self):
        self.listeners = {}     # map listening {(q2qid, protocol name):(protocol
                                # factory, protocol description)}
        self.pumps = []         # a list of IOPumps that we have to flush
        self.calls = []
        self.time = 0

    def callLater(self, s, f, *a, **k):
        # XXX TODO: return canceller
        assert f is not None
        tup = (self.time + s, f, a, k)
        self.calls.append(tup)
        self.calls.sort()
        return FakeDelayedCall(self, tup)

    def flush(self, debug=False):
        result = True
        while result:
            self.time += 1
            result = False
            for x in range(2):
                # run twice so that timed functions can interact with I/O
                for pump in self.pumps:
                    if pump.flush(debug):
                        result = True
                if debug:
                    print 'iteration finished.  continuing?', result
                c = self.calls
                self.calls = []
                for s, f, a, k in c:
                    if debug:
                        print 'timed event', s, f, a, k
                    f(*a,**k)
        return result

    def listenQ2Q(self, fromAddress, protocolsToFactories, serverDescription):
        for pname, pfact in protocolsToFactories.items():
            self.listeners[fromAddress, pname] = pfact, serverDescription
        return defer.succeed(None)

    def connectQ2Q(self, fromAddress, toAddress,
                   protocolName, protocolFactory,
                   chooser=lambda x: x and [x[0]]):
        # XXX update this when q2q is updated to return a connector rather than
        # a Deferred.

        # XXX this isn't really dealing with the multiple-connectors use case
        # now.  sigma doesn't need this functionality, but we will need to
        # update this class to do it properly before using it to test other Q2Q
        # code.

        listener, description = self.listeners.get((toAddress, protocolName))
        if listener is None:
            print 'void listener', fromAddress, toAddress, self.listeners, self.listener
            reason = Failure(KeyError())
            protocolFactory.clientConnectionFailed(None, reason)
            return defer.fail(reason)
        else:
            cio = StringIO()
            sio = StringIO()

            server = listener.buildProtocol(fromAddress)
            client = protocolFactory.buildProtocol(toAddress)

            servertpt = FakeQ2QTransport(sio, toAddress, fromAddress)
            clienttpt = FakeQ2QTransport(cio, fromAddress, toAddress)

            server.makeConnection(servertpt)
            client.makeConnection(clienttpt)
            pump = IOPump(client, server, cio, sio, False)
            self.pumps.append(pump)

            return defer.succeed(client)


sender = Q2QAddress("sending-data.net", "sender")
receiver = Q2QAddress("receiving-data.org", "receiver")

class TestBase(unittest.TestCase):
    def setUp(self):
        self.realChunkSize = sigma.CHUNK_SIZE
        sigma.CHUNK_SIZE = 100
        svc = self.service = FakeQ2QService()
        fname = self.mktemp()

        sf = self.sfile = FilePath(fname)
        if not sf.parent().isdir():
            sf.parent().makedirs()
        sf.open('w').write(TEST_DATA)
        self.senderNexus = sigma.Nexus(svc, sender,
                                       sigma.BaseNexusUI(self.mktemp()),
                                       svc.callLater)

    def tearDown(self):
        self.senderNexus.stopService()
        sigma.CHUNK_SIZE = self.realChunkSize


class BasicTransferTest(TestBase):
    def setUp(self):
        TestBase.setUp(self)
        self.stoppers = []
        self.receiverNexus = sigma.Nexus(self.service, receiver,
                                         sigma.BaseNexusUI(self.mktemp()),
                                         self.service.callLater)
        self.stoppers.append(self.receiverNexus)


    def tearDown(self):
        TestBase.tearDown(self)
        for stopper in self.stoppers:
            stopper.stopService()


    def testOneSenderOneRecipient(self):
        self.senderNexus.push(self.sfile, 'TEST->TEST', [receiver])
        self.service.flush()
        peerThingyoes = childrenOf(self.receiverNexus.ui.basepath)
        self.assertEquals(len(peerThingyoes), 1)
        rfiles = childrenOf(peerThingyoes[0])
        self.assertEquals(len(rfiles), 1)
        rfile = rfiles[0]
        rfdata = rfile.open().read()
        self.assertEquals(len(rfdata),
                          len(TEST_DATA))
        self.assertEquals(rfdata, TEST_DATA,
                          "file values unequal")

    def testOneSenderManyRecipients(self):
        raddresses = [Q2QAddress("receiving-data.org", "receiver%d" % (x,))
                      for x in range(10)]

        nexi = [sigma.Nexus(self.service,
                            radr,
                            sigma.BaseNexusUI(self.mktemp()),
                            self.service.callLater) for radr in raddresses]

        self.stoppers.extend(nexi)

        self.senderNexus.push(self.sfile, 'TEST->TEST', raddresses)
        self.service.flush()

        receivedIntroductions = 0

        for nexium in nexi:
            receivedIntroductions += nexium.ui.receivedIntroductions
        self.failUnless(receivedIntroductions > 1)

        for nexium in nexi:
            peerFiles = childrenOf(nexium.ui.basepath)
            self.assertEquals(len(peerFiles), 1)
            rfiles = childrenOf(peerFiles[0])
            self.assertEquals(len(rfiles), 1, rfiles)
            rfile = rfiles[0]
            self.assertEquals(rfile.open().read(),
                              TEST_DATA,
                              "file value mismatch")


def childrenOf(x):
    # this should be a part of FilePath, but hey
    return map(x.child, x.listdir())



# Copyright 2005 Divmod, Inc.  See LICENSE file for details
# -*- test-case-name: vertex.test.test_q2q.TCPConnection.testSendingFiles -*-

"""
Connect between two endpoints using a message-based protocol to exchange
messages lazily in response to UI events, caching the protocol as necessary.
Using connection-oriented protocols, you will most likely not want to use this
class - you might end up retrieving a cached connection in the middle of a
chunk of data being sent.  For the purposes of this distinction, a
'message-oriented' protocol is one which has an API which either:

  a) writes only whole messages to its transport so there is never an
     opportunity to insert data into the middle of a message, or

  b) provides an API on the Protocol instance for queuing whole messages such
     that if partial messages are sent, calling the API multiple times will
     queue them internally so that clients do not need to care whether the
     connection is made or not.

It is worth noting that all Juice-derived protocols meet constraint (b).
"""

from zope.interface import implements

from twisted.internet.defer import maybeDeferred, DeferredList, Deferred
from twisted.internet.main import CONNECTION_LOST
from twisted.internet import interfaces
from twisted.internet.protocol import ClientFactory


class ConnectionCache:
    def __init__(self):
        """
        """
        # map (fromAddress, toAddress, protoName): protocol instance
        self.cachedConnections = {}
        # map (fromAddress, toAddress, protoName): list of Deferreds
        self.inProgress = {}

    def connectCached(self, endpoint, protocolFactory,
                      extraWork=lambda x: x,
                      extraHash=None):
        """See module docstring
        """
        key = endpoint, extraHash
        D = Deferred()
        if key in self.cachedConnections:
            D.callback(self.cachedConnections[key])
        elif key in self.inProgress:
            self.inProgress[key].append(D)
        else:
            self.inProgress[key] = [D]
            endpoint.connect(
                _CachingClientFactory(
                    self, key, protocolFactory,
                    extraWork))
        return D

    def cacheUnrequested(self, endpoint, extraHash, protocol):
        self.connectionMadeForKey((endpoint, extraHash), protocol)

    def connectionMadeForKey(self, key, protocol):
        deferreds = self.inProgress.pop(key, [])
        self.cachedConnections[key] = protocol
        for d in deferreds:
            d.callback(protocol)

    def connectionLostForKey(self, key):
        if key in self.cachedConnections:
            del self.cachedConnections[key]

    def connectionFailedForKey(self, key, reason):
        deferreds = self.inProgress.pop(key)
        for d in deferreds:
            d.errback(reason)

    def shutdown(self):
        return DeferredList(
            [maybeDeferred(p.transport.loseConnection)
             for p in self.cachedConnections.values()])


class _CachingClientFactory(ClientFactory):
    debug = False

    def __init__(self, cache, key, subFactory, extraWork):
        """
        @param cache: a Q2QService

        @param key: a 2-tuple of (endpoint, extra) that represents what
        connections coming from this factory are for.

        @param subFactory: a ClientFactory which I forward methods to.

        @param extraWork: extraWork(proto) -> Deferred which fires when the
        connection has been prepared sufficiently to be used by subsequent
        connections and can be counted as a success.
        """

        self.cache = cache
        self.key = key
        self.subFactory = subFactory
        self.finishedExtraWork = False
        self.extraWork = extraWork

    lostAsFailReason = CONNECTION_LOST

    def clientConnectionMade(self, protocol):
        def success(reason):
            self.cache.connectionMadeForKey(self.key, protocol)
            self.finishedExtraWork = True
            return protocol

        def failed(reason):
            self.lostAsFailReason = reason
            protocol.transport.loseConnection()
            return reason
        maybeDeferred(self.extraWork, protocol).addCallbacks(
            success, failed)

    def clientConnectionLost(self, connector, reason):
        if self.finishedExtraWork:
            self.cache.connectionLostForKey(self.key)
        else:
            self.cache.connectionFailedForKey(self.key,
                                              self.lostAsFailReason)
        self.subFactory.clientConnectionLost(connector, reason)

    def clientConnectionFailed(self, connector, reason):
        self.cache.connectionFailedForKey(self.key, reason)
        self.subFactory.clientConnectionFailed(connector, reason)

    def buildProtocol(self, addr):
        return _CachingTransportShim(self, self.subFactory.buildProtocol(addr))


class _CachingTransportShim:
    disconnecting = property(lambda self: self.transport.disconnecting)

    implements(interfaces.IProtocol)

    def __init__(self, factory, protocol):
        self.factory = factory
        self.protocol = protocol

        # IProtocol
        self.dataReceived = protocol.dataReceived
        self.connectionLost = protocol.connectionLost


    def makeConnection(self, transport):
        self.transport = transport
        self.protocol.makeConnection(transport)
        self.factory.clientConnectionMade(self.protocol)


    def __repr__(self):
        return 'Q2Q-Cached<%r, %r>' % (self.transport,
                                       self.protocol)





from distutils.core import setup

setup(
    name="Vertex",
    version="0.1",
    maintainer="Divmod, Inc.",
    maintainer_email="support@divmod.org",
    url="http://divmod.org/trac/wiki/VertexProject",
    license="MIT",
    platforms=["any"],
    description="A Q2Q protocol implementation",
    classifiers=(
        "Intended Audience :: Developers",
        "Programming Language :: Python",
        "Development Status :: 2 - Pre-Alpha",
        "Topic :: Internet"),

    packages=['vertex',
              'vertex.scripts',
              'vertex.test'],

    scripts=['bin/vertex', 'bin/certcreate'])



# Copyright 2005 Divmod, Inc.  See LICENSE file for details
# -*- vertex.test.test_q2q.UDPConnection -*-

import os
from cStringIO import StringIO

from twisted.trial import unittest, util
from twisted.application import service
from twisted.internet import reactor, protocol, defer
from twisted.protocols import basic
from twisted.python import log
from twisted.python import failure
from twisted.cred import checkers
from twisted.internet.error import ConnectionDone
# from twisted.internet.main import CONNECTION_DONE

from zope.interface import implements
from twisted.internet.interfaces import IResolverSimple

from vertex import q2q, q2qclient, sslverify, juice
def noResources(*a):
    return []

class FakeConnectTCP:
    implements(IResolverSimple)

    def __init__(self, connectTCP):
        self._connectTCP = connectTCP
        self.hostPortToHostPort = {}
        self.hostToLocalHost = {}
        self.counter = 1

    def addHostPort(self, hostname, fakePortNumber, realPortNumber):
        if hostname in self.hostToLocalHost:
            localIP = self.hostToLocalHost[hostname]
        else:
            localIP = '127.0.0.%d' % (self.counter,)
            self.counter += 1
            self.hostToLocalHost[hostname] = localIP

        self.hostPortToHostPort[(localIP, fakePortNumber)] = (localIP, realPortNumber)
        self.hostPortToHostPort[(hostname, fakePortNumber)] = (hostname, realPortNumber)

    def connectTCP(self, host, port, *args, **kw):
        localhost, localport = self.hostPortToHostPort.get((host,port), (host, port))
        return self._connectTCP(localhost, localport, *args, **kw)

    def getHostSync(self,name):
        result = self.hostToLocalHost[name]
        return result

    def getHostByName(self, name, timeout):
        return defer.maybeDeferred(self.getHostSync, name)

def runOneDeferred(d):
    L = []
    d.addBoth(L.append)
    reactor.callLater(0, d.addCallback, lambda ign: reactor.crash())
    reactor.run()
    if L:
        if isinstance(L[0], failure.Failure):
            L[0].trap()
        return L[0]
    raise unittest.FailTest("Keyboard Interrupt")

class Utility(unittest.TestCase):
    def testServiceInitialization(self):
        svc = q2q.Q2QService(noResources)
        svc.certificateStorage.addPrivateCertificate("test.domain")

        cert = svc.certificateStorage.getPrivateCertificate("test.domain")
        self.failUnless(cert.getPublicKey().matches(cert.privateKey))

class OneTrickPony(juice.Juice):
    def juice_TRICK(self, box):
        return juice.QuitBox(tricked='True')

class OneTrickPonyServerFactory(juice.JuiceServerFactory):
    protocol = OneTrickPony

class OneTrickPonyClient(juice.Juice):
    def connectionMade(self):
        self.sendCommand('trick').chainDeferred(self.factory.ponged)

class OneTrickPonyClientFactory(juice.JuiceClientFactory):
    protocol = OneTrickPonyClient

    def __init__(self, ponged):
        self.ponged = ponged

    def buildProtocol(self, addr):
        result = juice.JuiceClientFactory.buildProtocol(self, addr)
        self.proto = result
        return result

    def clientConnectionFailed(self, connector, reason):
        self.ponged.errback(reason)


class DataEater(protocol.Protocol):
    def __init__(self):
        self.waiters = []
        self.data = []
        self.count = 0

    def dataReceived(self, data):
        if not data:
            raise RuntimeError("Empty string delivered to DataEater")
        self.data.append(data)
        self.count += len(data)
        for count, waiter in self.waiters[:]:
            if self.count >= count:
                waiter.callback(self.count)

    def removeD(self, result, size, d):
        # XXX done as a callback because 1.3 util.wait actually calls a
        # callback on the deferred
        self.waiters.remove((size, d))
        return result

    def waitForCount(self, size):
        D = defer.Deferred()
        self.waiters.append((size, D))
        self.waiters.sort()
        self.waiters.reverse()
        return D.addBoth(self.removeD, size, D)

    def buildProtocol(self, addr):
        return self

class DataFeeder(protocol.Protocol):
    def __init__(self, fobj):
        self.fobj = fobj

    def clientConnectionFailed(self, connector, reason):
        log.msg("DataFeeder client connection failed:")
        log.err(reason)

    def clientConnectionLost(self, connector, reason):
        pass

    def connectionMade(self):
        basic.FileSender().beginFileTransfer(self.fobj, self.transport)

    def buildProtocol(self, addr):
        return self

class StreamingDataFeeder(protocol.Protocol):
    DELAY = 0.01
    CHUNK = 1024
    paused = False
    pauseCount = 0
    resumeCount = 0
    stopCount = 0
    outCount = 0
    call = None

    def __init__(self, infile):
        self.file = infile

    def clientConnectionFailed(sef, connector, reason):
        log.msg("StreamingDataFeeder client connection failed:")
        log.err(reason)

    def clientConnectionLost(self, connector, reason):
        pass

    def connectionMade(self):
        self.nextChunk = self.file.read(self.CHUNK)
        self.transport.registerProducer(self, True)
        self.call = reactor.callLater(self.DELAY, self._keepGoing)

    def _keepGoing(self):
        self.call = None
        if self.paused:
            return
        chunk = self.nextChunk
        self.nextChunk = self.file.read(self.CHUNK)
        self.outCount += len(chunk)
        if chunk:
            self.transport.write(chunk)
        if self.nextChunk:
            self.call = reactor.callLater(self.DELAY, self._keepGoing)


    def pauseProducing(self):
        self.paused = True
        self.pauseCount += 1
        if self.call is not None:
            self.cancelMe()

    def resumeProducing(self):
        self.paused = False
        if self.call is not None:
            self.cancelMe()
        self.call = reactor.callLater(self.DELAY, self._keepGoing)
        self.resumeCount += 1

    def cancelMe(self):
        self.call.cancel()
        self.call = None

    def stopProducing(self):
        self.paused = True
        self.stopCount += 1
        if self.call is not None:
            self.cancelMe()

    def buildProtocol(self, addr):
        return self


class ErroneousClientError(Exception):
    pass

class EngenderError(juice.Command):
    commandName = 'Engender-Error'

class Break(juice.Command):
    commandName = 'Break'

class Flag(juice.Command):
    commandName = 'Flag'

class Erroneous(juice.Juice):
    flag = False
    def command_BREAK(self):
        raise ErroneousClientError("Zoop")
    command_BREAK.command = Break

    def command_ENGENDER_ERROR(self):
        return Break().do(self)
    command_ENGENDER_ERROR.command = EngenderError

    def command_FLAG(self):
        self.flag = True
    command_FLAG.command = Flag

class ErroneousServerFactory(juice.JuiceServerFactory):
    protocol = Erroneous

class ErroneousClientFactory(juice.JuiceClientFactory):
    protocol = Erroneous

class Greet(juice.Command):
    commandName = 'Greet'

class Greeter(juice.Juice, protocol.ServerFactory, protocol.ClientFactory):
    def __init__(self, issueGreeting, startupD):
        juice.Juice.__init__(self, issueGreeting)
        self.startupD = startupD

    def buildProtocol(self, addr):
        return self

    def connectionMade(self):
        Greet().do(self).chainDeferred(self.startupD)

    def command_GREET(self):
        self.greeted = True
        return dict()
    command_GREET.command = Greet

class Q2QConnectionTestCase(unittest.TestCase):
    streamer = None

    fromResource = 'clientResource'
    toResource = 'serverResource'

    fromDomain = 'origin.domain.example.com'
    spoofedDomain = 'spoofed.domain.example.com'
    toDomain = 'destination.domain.example.org'

    userReverseDNS = 'i.watch.too.much.tv'
    inboundTCPPortnum = 0
    udpEnabled = False
    virtualEnabled = False

    def _makeQ2QService(self, certificateEntity, pff=None):
        svc = q2q.Q2QService(pff, q2qPortnum=0,
                             inboundTCPPortnum=self.inboundTCPPortnum)
        svc.udpEnabled = self.udpEnabled
        svc.virtualEnabled = self.virtualEnabled
        if '@' not in certificateEntity:
            svc.certificateStorage.addPrivateCertificate(certificateEntity)
        svc.debugName = certificateEntity
        return svc


    def _addQ2QProtocol(self, name, factory):
        resourceKey = (self.fromAddress,
                       self.toAddress, name)
        self.resourceMap[resourceKey] = factory

    def protocolFactoryFactory(self, *key):
        if key in self.resourceMap:
            return [(self.resourceMap[key], 'test-description')]
        return []


    def setUp(self):
        log.ignoreErrors(ConnectionDone)
        self.fromAddress = q2q.Q2QAddress(self.fromDomain, self.fromResource)
        self.toAddress = q2q.Q2QAddress(self.toDomain, self.toResource)

        # A mapping of host names to port numbers Our connectTCP will always
        # connect to 127.0.0.1 and on a port which is a value in this
        # dictionary.
        fakeDNS = FakeConnectTCP(reactor.connectTCP)
        reactor.connectTCP = fakeDNS.connectTCP

        # ALSO WE MUST DO OTHER SIMILAR THINGS
        self._oldResolver = reactor.resolver
        reactor.installResolver(fakeDNS)

        # Set up a know-nothing service object for the client half of the
        # conversation.
        self.serverService2 = self._makeQ2QService(self.fromDomain, noResources)

        # Do likewise for the server half of the conversation.  Also, allow
        # test methods to set up some trivial resources which we can attempt to
        # access from the client.
        self.resourceMap = {}
        self.serverService = self._makeQ2QService(self.toDomain,
                                                  self.protocolFactoryFactory)

        self.msvc = service.MultiService()
        self.serverService2.setServiceParent(self.msvc)
        self.serverService.setServiceParent(self.msvc)

        # Let the kernel allocate a random port for each of these service's listeners
        self.msvc.startService()

        fakeDNS.addHostPort(
            self.fromDomain, 8788,
            self.serverService2.q2qPort.getHost().port)

        fakeDNS.addHostPort(
            self.toDomain, 8788,
            self.serverService.q2qPort.getHost().port)

        self._addQ2QProtocol('pony', OneTrickPonyServerFactory())

        self.dataEater = DataEater()
        self._addQ2QProtocol('eat', self.dataEater)

        self._addQ2QProtocol('error', ErroneousServerFactory())

    def tearDown(self):
        reactor.installResolver(self._oldResolver)
        del reactor.connectTCP
        return self.msvc.stopService()
        # log.clearIgnores()

class ConnectionTestMixin:

    def testConnectWithIntroduction(self):
        ponged = defer.Deferred()
        self.serverService2.connectQ2Q(self.fromAddress,
                                      self.toAddress,
                                      'pony',
                                      OneTrickPonyClientFactory(ponged))
        return ponged.addCallback(lambda answerBox: self.failUnless('tricked' in answerBox))

    def addClientService(self, toAddress, secret, serverService):
        return self._addClientService(
            toAddress.resource, secret, serverService, toAddress.domain)

    def _addClientService(self, username,
                          privateSecret, serverService,
                          serverDomain):
        svc = self._makeQ2QService(username + '@' + serverDomain)
        serverService.certificateStorage.addUser(serverDomain,
                                                 username,
                                                 privateSecret)
        svc.setServiceParent(self.msvc)
        return svc.authorize(q2q.Q2QAddress(serverDomain, username),
                             privateSecret).addCallback(lambda x: svc)


    def testListening(self):

        self.clientServerService = util.wait(self.addClientService(
                self.toAddress, 'aaaa', self.serverService))


        ponyFactory = OneTrickPonyServerFactory()
        ponged = defer.Deferred()
        util.wait(self.clientServerService.listenQ2Q(self.toAddress,
                                                {'pony2': ponyFactory},
                                                'ponies suck'))


        self.clientClientService = util.wait(self.addClientService(
                self.fromAddress, 'bbbb', self.serverService2))
        otpcf = OneTrickPonyClientFactory(ponged)
        self.clientClientService.connectQ2Q(self.fromAddress,
                                            self.toAddress,
                                            'pony2',
                                            otpcf)

        answerBox = util.wait(ponged)

        T = otpcf.proto.transport
        self.assertEquals(T.getQ2QPeer(), self.toAddress)
        self.assertEquals(T.getQ2QHost(), self.fromAddress)
        self.failUnless('tricked' in answerBox)

    def testChooserGetsThreeChoices(self):
        ponyFactory = OneTrickPonyServerFactory()
        self.testListening()

        # print 'dang yo'

        self.clientServerService2 = util.wait(self.addClientService(
                self.toAddress, 'aaaa', self.serverService))

        # print 'ultra frack'

        util.wait(self.clientServerService2.listenQ2Q(self.toAddress,
                                                      {'pony': ponyFactory},
                                                      'ponies are weird'))

        # print 'frack'

        ponged = defer.Deferred()
        util.wait(self.clientServerService.listenQ2Q(self.toAddress,
                                                     {'pony': ponyFactory},
                                                     'ponies rule'))

        # print 'WAITING IS DONE'

        expectedList = ['ponies rule', 'ponies are weird', 'test-description']
        def chooser(servers):
            self.failUnlessEqual(len(servers), 3)
            for server in servers:
                expectedList.remove(server['description'])
                if server['description'] == 'ponies rule':
                    self.assertEquals(
                        self.clientServerService.certificateStorage.getPrivateCertificate(str(self.toAddress)),
                        server['certificate'])
                    yield server

        util.wait(self.clientClientService.connectQ2Q(
                self.fromAddress,
                self.toAddress,
                'pony',
                OneTrickPonyClientFactory(ponged),
                chooser=chooser))

        self.failUnlessEqual(expectedList, [])

    def testTwoGreetings(self):
        d1 = defer.Deferred()
        d2 = defer.Deferred()
        client = Greeter(False, d1)
        server = Greeter(True, d2)
        self._addQ2QProtocol('greet', server)
        self.serverService2.connectQ2Q(self.fromAddress,
                                      self.toAddress,
                                      'greet',
                                      client)
        util.wait(defer.DeferredList([d1, d2]))
        self.failUnless(client.greeted)
        self.failUnless(server.greeted)


    def testSendingFiles(self):
        SIZE = 1024 * 500
        self.streamer = StreamingDataFeeder(StringIO('x' * SIZE))
        self.streamer.CHUNK = 8192
        util.wait(self.serverService2.connectQ2Q(self.fromAddress,
                                                 self.toAddress, 'eat',
                                                 DataFeeder(StringIO('y' * SIZE))))
        util.wait(self.serverService2.connectQ2Q(self.fromAddress,
                                                 self.toAddress, 'eat',
                                                 self.streamer))

        # self.assertEquals( len(self.serverService.liveConnections), 1)
        # XXX currently there are 2 connections but there should only be 1: the
        # connection cache is busted, need a separate test for that
        for liveConnection in self.serverService.iterconnections():
            liveConnection.transport.pauseProducing()
        wfc = self.dataEater.waitForCount(SIZE * 2)
        self.assertRaises(defer.TimeoutError, util.wait, wfc, timeout=3)
        for liveConnection in self.serverService.iterconnections():
            liveConnection.transport.resumeProducing()
        util.wait(self.dataEater.waitForCount(SIZE * 2), 30)
        self.failUnless(self.streamer.pauseCount > 0)
        self.failUnless(self.streamer.resumeCount > 0)

    def testBadIssuerOnSelfSignedCert(self):
        x = self.testConnectWithIntroduction()
        def actualTest(result):
            ponged = defer.Deferred()
            signer = self.serverService2.certificateStorage.getPrivateCertificate(
                self.fromDomain).privateKey
            req = signer.requestObject(sslverify.DN(commonName=self.toDomain))
            sreq = signer.signRequestObject(sslverify.DN(commonName=self.fromDomain),
                                                       req, 12345)
            selfSignedLie = sslverify.PrivateCertificate.fromCertificateAndKeyPair(
                sreq, signer)
            self.serverService2.connectQ2Q(self.fromAddress,
                                          self.toAddress,
                                          'pony',
                                          OneTrickPonyClientFactory(ponged),
                                          selfSignedLie,
                                          fakeFromDomain=self.toDomain).addErrback(
                lambda e: e.trap(sslverify.VerifyError))

            return self.assertFailure(ponged, sslverify.VerifyError)
        return x.addCallback(actualTest)


    def testBadCertRequestSubject(self):
        kp = sslverify.KeyPair.generate()
        subject = sslverify.DN(commonName='HACKERX',
                               localityName='INTERNETANIA')
        reqobj = kp.requestObject(subject)

        fakereq = kp.requestObject(subject)
        ssigned = kp.signRequestObject(subject, fakereq, 1)
        certpair = sslverify.PrivateCertificate.fromCertificateAndKeyPair
        fakecert = certpair(ssigned, kp)
        apc = self.serverService2.certificateStorage.addPrivateCertificate

        def _2(secured):
            D = q2q.Sign(certificate_request=reqobj,
                         password='itdoesntmatter').do(secured)
            def _1(dcert):
                cert = dcert['certificate']
                privcert = certpair(cert, kp)
                apc(str(self.fromAddress), privcert)
            return D.addCallback(_1)

        d = self.serverService2.getSecureConnection(
            self.fromAddress, self.fromAddress.domainAddress(), authorize=False,
            usePrivateCertificate=fakecert,
            ).addCallback(_2)

        def unexpectedSuccess(result):
            self.fail("Expected BadCertificateRequest, got %r" % (result,))
        def expectedFailure(err):
            err.trap(q2q.BadCertificateRequest)
        d.addCallbacks(unexpectedSuccess, expectedFailure)
        util.wait(d)

    def testClientSideUnhandledException(self):
        d = self.serverService2.connectQ2Q(
            self.fromAddress, self.toAddress, 'error',
            ErroneousClientFactory())
        def connected(proto):
            return EngenderError().do(proto)
        def exploded(err):
            err.trap(ConnectionDone)
        d.addCallback(connected)
        d.addCallbacks(self.successIsFailure, exploded)
        d.addCallback(lambda ign: log.flushErrors(ErroneousClientError))
        util.wait(d)

    def successIsFailure(self, success):
        self.fail()

    def testTwoBadWrites(self):
        d = self.serverService2.connectQ2Q(
            self.fromAddress, self.toAddress, 'error',
            ErroneousClientFactory())

        def connected(proto):
            def trapit(what):
                what.trap(juice.UnhandledRemoteJuiceError)
            Break().do(proto).addCallbacks(self.successIsFailure, trapit)
            return Flag().do(proto)
        def err(x):
            x.trap(ConnectionDone)

        d.addCallback(connected)
        d.addCallbacks(self.successIsFailure, err)
        d.addCallback(lambda ign: log.flushErrors(ErroneousClientError))
        util.wait(d)




class VirtualConnection(Q2QConnectionTestCase, ConnectionTestMixin):
    inboundTCPPortnum = None
    udpEnabled = False
    virtualEnabled = True

    def testListening(self):
        pass

    def testChooserGetsThreeChoices(self):
        pass

    testListening.skip = 'virtual port forwarding not implemented'
    testChooserGetsThreeChoices.skip = 'cant do this without testListening'

class UDPConnection(Q2QConnectionTestCase, ConnectionTestMixin):
    # skip = 'yep'
    inboundTCPPortnum = None
    udpEnabled = True
    virtualEnabled = False

class TCPConnection(Q2QConnectionTestCase, ConnectionTestMixin):
    inboundTCPPortnum = 0
    udpEnabled = False
    virtualEnabled = False

class TestProtocol(juice.Juice):
    def juice_GETADDRESSINFO(self, request):
        h = self.transport.getHost()
        p = self.transport.getPeer()
        return juice.Box(
            Host_Resource=h.resource,
            Host_Domain=h.domain,
            Peer_Resource=p.resource,
            Peer_Domain=p.domain)

class TestServerFactory(juice.JuiceClientFactory):
    protocol = TestProtocol

# A special treat for Glyph to enjoy later.

def _findService(svc, matcher):
    try:
        truth = matcher(svc)
    except:
        log.err()
        truth = False
    if truth:
        yield svc
    try:
        i = iter(svc)
    except:
        # print 'Not iterable:', svc
        return
    for subsvc in i:
        for blah in _findService(subsvc, matcher):
            yield blah

# class LiveServerMixin:
#     serverDomain = 'test.domain.example.com'

#     def jackDNS(self, *info):
#         self.fakeDNS = FakeConnectTCP(reactor.connectTCP)
#         reactor.connectTCP = self.fakeDNS.connectTCP
#         self._oldResolver = reactor.resolver
#         reactor.installResolver(self.fakeDNS)

#         for (hostname, oldport, newport) in info:
#             self.fakeDNS.addHostPort(hostname, oldport, newport)

#     def unjackDNS(self):
#         del reactor.connectTCP
#         reactor.installResolver(self._oldResolver)

# class AuthorizeTestCase(unittest.TestCase, LiveServerMixin):
#     authUser = 'authtestuser'
#     authPass = 'p4ssw0rd'

#     def setUp(self):
#         self.testDir, serverService = self.deploy()
#         self.serverService = service.IServiceCollection(serverService)
#         self.serverService.startService()
#         self.clientDir = os.path.join(self.testDir, 'client')
#         self.clientService = q2qclient.ClientQ2QService(self.clientDir)

#         q2qPort = self.getQ2QService().q2qPort.getHost().port
#         self.jackDNS((self.serverDomain, 8788, q2qPort))

#     def tearDown(self):
#         self.unjackDNS()
#         util.wait(self.serverService.stopService())
#         util.wait(self.clientService.stopService())

#     def testAuthorize(self):
#         self.createUser(self.authUser, self.authPass)

#         d = self.clientService.authorize(
#             q2q.Q2QAddress(self.serverDomain, self.authUser),
#             self.authPass)

#         result = runOneDeferred(d)

#         self.failUnless(os.path.exists(os.path.join(self.clientDir, 'public')))
#         self.failUnless(os.path.exists(os.path.join(self.clientDir, 'public', self.serverDomain + '.pem')))
#         self.failUnless(os.path.exists(os.path.join(self.clientDir, 'private')))
#         self.failUnless(os.path.exists(os.path.join(self.clientDir, 'private', self.authUser + '@' + self.serverDomain + '.pem')))



# Copyright 2005 Divmod, Inc.  See LICENSE file for details

import sys

from twisted.python import usage

from vertex import sslverify

class Options(usage.Options):
    optParameters = [
        ["country", "C", "US", None],
        ["state", "s", "New York", None],
        ["city", "c", "New York", None],
        ["organization", "o", "Divmod LLC", None],
        ["unit", "u", "Security", None],
        ["hostname", "h", "divmod.com", None],
        ["email", "e", "support@divmod.org", None],

        ["filename", "f", "server.pem", "Name of the file to which to write the PEM."]]

def createSSLCertificate(opts):
    sslopt = {}
    for x, y in (('country','C'),
                 ('state', 'ST'),
                 ('city', 'L'),
                 ('organization', 'O'),
                 ('unit', 'OU'),
                 ('hostname', 'CN'),
                 ('email','emailAddress')):
        sslopt[y] = opts[x]
    ssc = sslverify.KeyPair.generate().selfSignedCert(1,**sslopt)
    file(opts['filename'], 'w').write(ssc.dumpPEM())
    print 'Wrote SSL certificate:'
    print ssc.inspect()
    return ssc

def main(args=None):
    """
    Create a private key and a certificate and write them to a file.
    """
    if args is None:
        args = sys.argv[1:]

    o = Options()
    try:
        o.parseOptions(args)
    except usage.UsageError, e:
        raise SystemExit(str(e))
    else:
        return createSSLCertificate(o)
